use std::{env, ffi, ops, ptr};

use x11::xlib::{self, BadGC};

use anyhow::Result;

use super::{error, screen, window};

pub struct Display {
    pub(super) ptr: *mut xlib::Display,
    name: String,
    keyboard_grab: Option<GrabbablesModes>,
    pointer_grab: Option<GrabbablesModes>,
}

// for keyboard/pointer grabs so Display is less messy
#[derive(Debug)]
struct GrabbablesModes {
    keyboard_mode: GrabMode,
    pointer_mode: GrabMode,
}

#[derive(Debug)]
pub enum Grabbables {
    Keyboard,
    Pointer,
}

impl std::fmt::Display for Grabbables {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Grabbables::Keyboard => write!(f, "keyboard"),
            Grabbables::Pointer => write!(f, "pointer"),
        }
    }
}

#[derive(Clone, Copy, Debug)]
pub enum GrabMode {
    Sync,
    Async,
}

impl From<GrabMode> for i32 {
    fn from(v: GrabMode) -> Self {
        match v {
            GrabMode::Sync => xlib::GrabModeSync,
            GrabMode::Async => xlib::GrabModeAsync,
        }
    }
}

/// The event mode for XAllowEvents
pub enum EventMode {
    AsyncPointer, 
    SyncPointer, 
    AsyncKeyboard, 
    SyncKeyboard, 
    ReplayPointer, 
    ReplayKeyboard, 
    AsyncBoth,
    SyncBoth
}

impl From<EventMode> for i32 {
    fn from(event_mode: EventMode) -> Self {
        match event_mode {
            EventMode::AsyncPointer => xlib::AsyncPointer, 
            EventMode::SyncPointer => xlib::SyncPointer, 
            EventMode::AsyncKeyboard => xlib::AsyncKeyboard, 
            EventMode::SyncKeyboard => xlib::SyncKeyboard, 
            EventMode::ReplayPointer => xlib::ReplayPointer, 
            EventMode::ReplayKeyboard => xlib::ReplayKeyboard, 
            EventMode::AsyncBoth => xlib::AsyncBoth,
            EventMode::SyncBoth => xlib::SyncBoth,
        }
    }
}

impl Display {
    /// Call XOpenDisplay to open a connection to the X Server.
    /// If `display_name` is `None`, the value of the `DISPLAY` environment variable will be used.
    pub fn open(display_name: Option<String>) -> Result<Self> {
        let name = ffi::CString::new(if let Some(name) = display_name {
            name
        } else {
            env::var("DISPLAY")?
        })?;

        let name_ptr = name.as_bytes().as_ptr();

        // try to open display and get either display pointer or null
        let display_ptr = unsafe { xlib::XOpenDisplay(name_ptr as *const i8) };

        // if display is null, return an error, otherwise return instance successfully
        if display_ptr == ptr::null_mut::<xlib::_XDisplay>() {
            Err(error::XError::OpenDisplayError(name.into_string()?).into())
        } else {
            Ok(Self {
                ptr: display_ptr,
                name: name.into_string()?,
                pointer_grab: None,
                keyboard_grab: None,
            })
        }
    }

    /// Calls XFlush to flush the output buffer.
    pub fn flush(&self) {
        unsafe {
            xlib::XFlush(self.ptr);
        }
    }

    // TODO: Figure out how to properly handle errors
    /// Calls XSync to flush the output buffer and then wait until all events have been received and processed
    /// by the server.
    /// The `discard` parameter specifies, whether to discard all events in the queue.
    pub fn sync(&self, discard: bool) {
        unsafe {
            xlib::XSync(self.ptr, discard.into());
        }
    }

    /// Calls xlib::XDefaultScreen to get the default screen number referenced by Display::open.
    /// This should be used to retrieve the screen number in applications that'll only use a single
    /// screen.
    pub fn default_screen_nr(&self) -> i32 {
        unsafe { xlib::XDefaultScreen(self.ptr) }
    }

    /// Gets the default screen
    pub fn default_screen(&self) -> screen::Screen {
        screen::Screen {
            ptr: unsafe { xlib::XDefaultScreenOfDisplay(self.ptr) },
        }
    }

    /// Gets the root window of the default screen
    pub fn default_root_window(&self) -> window::Window {
        window::Window {
            wid: unsafe { xlib::XDefaultRootWindow(self.ptr) },
        }
    }

    /// Get the range of legal KeyCodes for a display.
    pub fn keycodes(&self) -> Result<ops::Range<i32>> {
        let (mut min, mut max) = (0, 0);

        if unsafe { xlib::XDisplayKeycodes(self.ptr, &mut min, &mut max) } == 0 {
            Err(error::XError::DisplayKeycodesError.into())
        } else {
            Ok(min..max)
        }
    }

    /// Performs an active Grab on the keyboard. Further key events are only reported to the
    /// grabbing client.
    pub fn grab_keyboard(
        &mut self,
        grab_window: window::Window,
        owner_events: bool,
        pointer_mode: GrabMode,
        keyboard_mode: GrabMode,
        time: xlib::Time,
    ) -> Result<()> {
        if let None = self.keyboard_grab {
            match unsafe {
                xlib::XGrabKeyboard(
                    self.ptr,
                    grab_window.wid,
                    owner_events.into(),
                    pointer_mode.into(),
                    keyboard_mode.into(),
                    time,
                )
            } {
                xlib::GrabSuccess => {
                    self.keyboard_grab = Some(GrabbablesModes {
                        keyboard_mode,
                        pointer_mode,
                    });

                    Ok(())
                }
                xlib::AlreadyGrabbed => 
                    Err(error::XError::XAlreadyGrabbed(Grabbables::Keyboard).into()),
                xlib::GrabInvalidTime => 
                    Err(error::XError::XGrabInvalidTime.into()),
                xlib::GrabNotViewable => 
                    Err(error::XError::XGrabNotViewable.into()),
                xlib::GrabFrozen => 
                    Err(error::XError::XGrabFrozen(Grabbables::Keyboard).into()),
                code => Err(error::XError::UnknownError(code).into()),
            }
        } else {
            Err(error::XError::AlreadyGrabbed(Grabbables::Keyboard).into())
        }
    }
    
    /// Ends the active keyboard grab.
    pub fn ungrab_keyboard(&self, time: xlib::Time) -> Result<()>{
        if let Some(_) = self.keyboard_grab {
            unsafe {
                xlib::XUngrabKeyboard(self.ptr, time);
                Ok(())
            }
        } else {
            Err(error::XError::NotGrabbed(Grabbables::Keyboard).into())
        }
    }

    pub fn allow_events(&self, event_mode: EventMode, time: xlib::Time) {
        unsafe {
            xlib::XAllowEvents(self.ptr, event_mode.into(), time);
        }
    }
}

impl Drop for Display {
    fn drop(&mut self) {
        if unsafe { xlib::XCloseDisplay(self.ptr) } == BadGC.into() {
            eprintln!("BadGC Error when closing display '{}'.", self.name);
        };
    }
}
