use std::{fmt, ops};

use x11::xlib;

use super::display;

/// Various errors to be used in this wrapper
#[derive(Debug)]
pub enum XError {
    OpenDisplayError(String),
    DisplayKeycodesError,
    InvalidKeycodeError(xlib::KeyCode, ops::Range<i32>),
    AlreadyGrabbed(display::Grabbables),
    NotGrabbed(display::Grabbables),
    XAlreadyGrabbed(display::Grabbables),
    XGrabFrozen(display::Grabbables),
    XGrabInvalidTime,
    XGrabNotViewable,
    XEventConversionError { event_type: i32 },
    UnknownError(i32),
}

impl std::fmt::Display for XError {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(
            f,
            "{}",
            match self {
                XError::OpenDisplayError(display_name) =>
                    format!("error when opening display '{}'", display_name),
                XError::DisplayKeycodesError => String::from("error when running XDisplayKeycodes"),
                XError::InvalidKeycodeError(code, range) =>
                    format!("keycode {} outside of range {:?}", code, range),
                XError::AlreadyGrabbed(thing_attempted_to_grab) => format!(
                    "this display already grabbed the {}",
                    thing_attempted_to_grab
                ),
                XError::NotGrabbed(thing_attempted_to_ungrab) => format!( 
                    "couldn't ungrab the {} because it wasn't grabbed", 
                    thing_attempted_to_ungrab
                ),
                XError::XAlreadyGrabbed(thing_attempted_to_grab) => format!(
                    "{} is already actively grabbed by another client",
                    thing_attempted_to_grab
                ),
                XError::XGrabFrozen(thing_attempted_to_grab) => format!(
                    "{} is frozen by an active grab of another client",
                    thing_attempted_to_grab
                ),
                XError::XGrabInvalidTime => String::from("invalid grab time"),
                XError::XGrabNotViewable => String::from("grab_window is not viewable"),
                XError::XEventConversionError { event_type } => format!(
                    "invalid event type: {}", 
                    event_type
                ),
                XError::UnknownError(code) => format!("unknown error code was returned: {}", code),
            }
        )
    }
}

impl std::error::Error for XError {}
